html, body {
  margin: 0;
  padding: 0;
}

body {
  min-width: 217px;
}

label, input[type="checkbox"] {
  margin-left: 10px;
}

input[type='text']#pwdlen {
  width: auto;
  padding: 10px;
  border: 0;
  box-sizing: border-box;
}

#search, input[type='text'] {
  width: 100%;
  padding: 10px;
  border: 0;
  box-sizing: border-box;
}

#search:focus, input[type='text']:focus {
  outline: 0;
}

#results {
  list-style-type: none;
  margin: 0;
  padding: 0;
}

#results:empty {
  display: none;
}

button.selected, button#close {
  width: auto;
  background-color: #fff;
  border-left: 1px solid black;
  border-right: 1px solid black;
  border-top: 1px solid black;
  box-shadow: none;
  text-align: left;
  padding: 10px;
}

button.inactive {
  width: auto;
  background-color: #aaa;
  border: 1px solid black;
  box-shadow: none;
  text-align: left;
  padding: 10px;
}

button.insert {
  width: 100%;
  color: blue;
  text-align: center;
}

button {
  width: 100%;
  background-color: #fff;
  border: 0;
  box-shadow: none;
  text-align: left;
  padding: 10px;
}

button:hover, button:focus, .focus button {
  outline: 0;
  background-color: #eee;
}

.hidden {
    display: none;
}
